'''matrixtools.sage - Created by Noud Aldenhoven (2014).'''

def SimplifyMatrix(mat, do_fac=True):
    '''SimplifyMatrix(mat, do_fac=True): Factorize and simplify every entry of matrix mat.'''
    
    def fsimplify(elm):
        elm = elm.full_simplify()
        if do_fac and elm != 0:
            elm = elm.factor()
        return elm

    return mat.apply_map(fsimplify)


def scalardiv(M1, M2, M, N):
    '''scalardiv(M1, M2, M, N): Divide two matrices M1 and M2 of dimension M x N entrywise.'''

    newm = matrix(SR, M, N)

    for i in range(M):
        for j in range(N):
            if M1[i,j] != 0:
                newm[i,j] = (M1[i,j] / M2[i,j]).full_simplify().factor()

    return newm


def antimatrix(F, N):
    '''antimatrix(F, N): Returns the anti-matrix J over field F of dimension N x N.'''

    newm = matrix(F, N, N)

    for i in range(N):
        newm[i, N-i-1] = 1

    return newm

def matrixcoefficients(M, x):
    '''matrixcoefficients(M, x): Split matrix M in matrix coefficients in x.'''

    height, width = M.dimensions()
    matrix_degree = 0
    for row in M:
        for elm in row:
            if elm.degree(x) > matrix_degree:
                matrix_degree = elm.degree(x)


    ret = [ [matrix(SR, width, height), i] for i in range(matrix_degree+1) ]
    for i in range(height):
        for j in range(width):
            all_coefficients = M[i,j].coefficients(x)
            for coeff, dgr in all_coefficients:
                ret[dgr][0][i,j] = coeff

    return ret

def matrixleadingcoefficient(M, x):
    try:
        return matrixcoefficients(M, x)[-1][0]
    except:
        height, width = M.dimensions()
        return identity_matrix(SR, height)

def monicmatrix(M, x):
    return SimplifyMatrix(M*matrixleadingcoefficient(M, x)**(-1))
