# Proof checking program for Matrix-valued orthogonal polynomials related 
# to the quantum analogue of (SU(2)xSU(2), diag)
#
# By Noud Aldenhoven, Erik Koelink and Pablo Roman
#
# Created July 2015

import itertools

load('qpolynomials.sage')
load('bhs.sage')
load('matrixtools.sage')
load('complextools.sage')
load('polynomials.sage')

x = var('x')
z = var('z')
q = var('q')

# The Clebsch-Gordan coefficients from Theorem 4.3
def cgc(l1, l2, l, i, j, k, verbose=False):
    if i - j != k or i < -l1 or i > l1 or j < -l2 or j > l2 or k < -l or k > l:
        return 0

    if l2 > l1:
        return cgc(l2, l1, l, j, i, -k)

    p = l1 - l2 - k
    if -2*l2 <= p and p < 0:
        # Case 1
        n = l1 - i
    elif 0 <= p:
        # Case 2 and 3
        n = l2 - j

    q = var('q')

    sign = (-1)**(l1+l2-l)

    if -2*l2 <= p and p < 0:
        # Case 1
        if verbose:
            print 'case 1'
        qpoch1 = qpoch(q**(-4*l2), q**2, 2*l2+p)
        qpoch2 = qpoch( 
            [q**(-4*l1), q**(-4*l1-4*l2-2), q**(-4*l2-2*p)],
            q**2, l1 + l2 - l)
        qpoch3 = qpoch(q**(-4*l1 - 4*l2), q**2, 2*l2 + p)
        qpoch4 = qpoch(
            [q**2, q**(-4*l1+2*p), q**(-4*l2)],
            q**2, l1 + l2 - l)
        line1 = sqrt((-1)**(l1 + l2 - l) * qpoch1 * qpoch2 / (qpoch3 * qpoch4))

        #print line1

        line2 = sqrt((1 - q**(-4*l-2)) \
            / (1 - q**(-4*l1 - 4*l2 - 2))) \
            * q**(-2*l1*(2*l2+p) + (l1+l2-l)*(2*l1+2*l2+p) \
            - binomial(l1 + l2 - l, 2))

        #print line2

        qpoch5 = qpoch(
            [q**(-4*l2-2*p), q**(-4*l1)],
            q**2, n)
        qpoch6 = qpoch(
            [q**2, q**(-2*p + 2)], q**2, n)
        #poly = bhs(
        #    [q**(-2*n), q**(-2*l1-2*l2+2*l), q**(-2*l1-2*l2-2*l-2)],
        #    [q**(-4*l1), q**(-4*l2-2*p)], q**2, q**2)
        poly = dual_q_Hahn(
            n, l1+l2-l, q**(-4*l1-2), q**(-4*l2-2), 2*l2+p, q**2
        )
        line3 = q**((2*l1 + 2*l2 + 1)*n) * sqrt(qpoch5 / qpoch6) * poly
        #print line3 
 
    elif 0 <= p and p <= 2*l1 - 2*l2:
        # Case 2
        if verbose:
            print 'case 2'

        qpoch1 = qpoch(q**(-4*l2 - 2*p), q**2, 2*l2)
        qpoch2 = qpoch(
            [q**(-4*l1 + 2*p), q**(-4*l1 - 4*l2 - 2), q**(-4*l2)], 
            q**2, 
            l1 + l2 - l)
        qpoch3 = qpoch(q**(-4*l1 - 4*l2), q**2, 2*l2)
        qpoch4 = qpoch(
            [q**2, q**(-4*l1), q**(-4*l2 - 2*p)], 
            q**2,
            l1 + l2 - l)
        line1 = sqrt((-1)**(l1 + l2 - l) * qpoch1 * qpoch2 / (qpoch3 * qpoch4))

        line2 = sqrt((1 - q**(-4*l - 2)) / (1 - q**(-4*l1 - 4*l2 - 2))) \
            * q**((l1+l2-l)*(2*l1+2*l2-p)+l2*(2*p-4*l1) \
            - binomial(l1 + l2 - l, 2))

        qpoch5 = qpoch([q**(-4*l2), q**(-4*l1 + 2*p)], q**2, n)
        qpoch6 = qpoch([q**2, q**(2*p + 2)], q**2, n)
        poly = dual_q_Hahn(
            n, 
            l1 + l2 - l, 
            q**(-4*l1 + 2*p - 2),
            q**(-4*l2 - 2*p - 2), 
            2*l2, 
            q**2)
        line3 = q**((2*l1 + 2*l2 + 1)*n) * sqrt(qpoch5 / qpoch6) * poly
    elif 2*l1 - 2*l2 < p and p <= 2*l1:
        # Case 3
        if verbose:
            print 'case 3'
        qpoch1 = qpoch(q**(-4*l1), q**2, 2*l1-p)
        qpoch2 = qpoch(
            [q**(-4*l2), q**(-4*l1-4*l2-2), q**(-4*l1+2*p)],
            q**2, l1 + l2 - l)
        qpoch3 = qpoch(q**(-4*l1 - 4*l2), q**2, 2*l1 - p)
        qpoch4 = qpoch(
            [q**2, q**(-4*l2-2*p), q**(-4*l1)],
            q**2, l1 + l2 - l)
        line1 = sqrt((-1)**(l1 + l2 - l) * qpoch1 * qpoch2 / (qpoch3 * qpoch4))

        line2 = sqrt((1 - q**(-4*l-2)) / (1 - q**(-4*l1 - 4*l2 - 2))) \
            * q**(-2*l2*(2*l1-p) + (2*l1+2*l2-p)*(l1+l2-l) - binomial(l1 + l2 - l, 2))

        qpoch5 = qpoch([q**(-4*l2), q**(-4*l1+2*p)], q**2, n)
        qpoch6 = qpoch([q**2, q**(2*p + 2)], q**2, n)
        poly = dual_q_Hahn(
            n, l1+l2-l, q**(-4*l2-2), q**(-4*l1-2), 2*l1-p, q**2
        )
        line3 = q**((2*l1 + 2*l2 + 1)*n) * sqrt(qpoch5 / qpoch6) * poly
        
    else:
        return 0

    return line1 * line2 * line3

def newcgc(l1, l2, l, i, j, k):
    return cgc(l2, l1, l, -j, -i, k)


###
# Spherical functions restricted to (K_1^{1/2}K_2^{1/2})^{\lambda} with action 
# A^{-1} from the left

def phi_vec(l, l1, l2):
    z = var('z')
    ret = []

    for n in srange(-l, l+1):
        elm = 0
        for n1 in srange(-l1, l1+1):
            n2 = n1 - n
            elm += newcgc(l1, l2, l, n1, n2, n)**2 * q**(n1+n2) * z**(-n1-n2)
        ret.append(elm)

    return ret

def Phi0(l):
    lmatrix = []
    
    for k in srange(-l, l+1):
        lmatrix.append(phi_vec(l, (l+k)/2, (l-k)/2))

    return matrix(lmatrix).transpose()

def Phi(l, n):
    lmatrix = []

    for k in srange(-l, l+1):
        lmatrix.append(phi_vec(l, (n + l + k)/2, (n + l - k)/2))

    return matrix(lmatrix).transpose()


###
# Constructing the polynomials from the spherical functions, Theorem 4.6 and
# Corollary 4.7.

def polynomial_from_spherical(l, n):
    return SimplifyMatrix(Phi(l,0)**(-1)*Phi(l,n)).apply_map(replace_z_x)

def monic_polynomial_from_spherical(l, n):
    return monicmatrix(polynomial_from_spherical(l, n), x) 


###
# The weight matrix of the matrix-valued polynomials, Theorem 4.8.

def weight_direct_cgc(l):
    rmatrix = matrix(SR, 2*l+1, 2*l+1)

    for k in range(2*l+1):
        for p in range(2*l+1):
            l1, l2 = k/2, l-k/2
            m1, m2 = p/2, l-p/2

            rsum = 0

            for n in srange(-l, l+1):
                for i1 in srange(-l1, l1+1):
                    for j1 in srange(-l2, l2+1):
                        for i2 in srange(-m1, m1+1):
                            for j2 in srange(-m2, m2+1):
                                rsum += newcgc(l1, l2, l, i1, j1, n)**2 * newcgc(m1, m2, l, i2, j2, n)**2 * q**(i1 + j1 + i2 + j2) * z**(-i1 - j1 + i2 + j2)

            rmatrix[k,p] = replace_z_x(rsum)

    return rmatrix

def weight_direct(l):
    def alpha(t, m, n):
        return q**(2*(2*l + 1)*n - n**2 - (4*l + 3)*t + t**2 - 2*l + m) \
            * (1 - q**(4*l + 2))/(1 - q**(2*m + 2)) \
            * qpoch(q**2, q**2, 2*l - n) * qpoch(q**2, q**2, n) / qpoch(q**2, q**2, 2*l) \
            * (-1)**(n-t) * qpoch(q**(2*m - 4*l), q**2, n - t) / qpoch(q**(2*m + 4), q**2, n - t) \
            * qpoch(q**(4*l + 4 - 2*t), q**2, t) / qpoch(q**2, q**2, t)


    ret = matrix(SR, 2*l+1, 2*l+1)
    for n1 in range(2*l+1):
        for m1 in range(n1, 2*l+1):
            for t in range(n1+1):
                ret[m1, n1] += alpha(t, m1, n1) * Chebychev2nd(m1+n1-2*t, x)
            if m1 != n1:
                ret[2*l-m1, 2*l-n1] = ret[m1, n1]

    return SimplifyMatrix(ret)


# Tools to check the orthogonality for the matrix-valued polynomials

def matrix_inproduct(W, P, Q):
    Mret = transpose(P)*W*Q

    return 2/pi * Mret.apply_map(lambda elm : integral(sqrt(1 - x**2)*elm, x, -1, 1))

def monic_Gn(l, m):
    mret = matrix(SR, 2*l+1)

    for i in range(2*l+1):
        mret[i,i] = 2**(-2*m) * q**(-2*l) * (1 - q**(4*l+2))**2 / ((1 - q**(2*m+2*i+2))*(1 - q**(4*l-2*i+2*m+2))) \
            * qpoch(q**2, q**2, m)**2 * qpoch(q**(4*l + 4), q**2, m)**2 / (qpoch(q**(2*i+2), q**2, m)**2 * qpoch(q**(4*l-2*i+2), q**2, m)**2)

    return mret

def Gn(l, m):
    mret = matrix(SR, 2*l+1)

    for i in range(2*l+1):
        mret[i,i] = q**(2*m - 2*l) * (1 - q**(4*l+2))**2 / ((1 - q**(2*m + 2*i + 2))*(1 - q**(4*l-2*i+2*m+2)))

    return mret

###
# Leading coefficient of Pn, 

def lc_from_pn(l, n):
    return SimplifyMatrix(matrixleadingcoefficient(polynomial_from_spherical(l, n), x))

def lc_direct(l, n):
    ret = matrix(SR, 2*l+1, 2*l+1)

    for i in range(2*l+1):
        ret[i,i] = 2**n*q**n * qpoch([q**(2*i + 2), q**(4*l - 2*i + 2)], q**2, n) / qpoch([q**2, q**(4*l + 4)], q**2, n)

    return ret


###
# Three-term recurrence relation Qn (monic), Theorem 4.11

def three_term_Xn(l, n):
    ret = matrix(SR, 2*l+1, 2*l+1)

    for i in range(2*l):
        ret[i,i+1] = q**(2*n+1)/2 * (1 - q**(2*i + 2))**2 / ((1 - q**(2*n + 2*i + 2))*(1 - q**(2*n + 2*i + 4)))

    for i in range(1, 2*l+1):
        ret[i,i-1] = q**(2*n+1)/2 * (1 - q**(4*l - 2*i + 2))**2 / ((1 - q**(4*l - 2*i + 2*n + 2))*(1 - q**(4*l - 2*i + 2*n + 4)))

    return ret

def three_term_Yn(l, n):
    ret = matrix(SR, 2*l+1, 2*l+1)

    for i in range(2*l+1):
        ret[i,i] = 4**(-1)*(1 - q**(2*n))**2*(1 - q**(4*l+2*n+2))**2 / ((1 - q**(2*i+2*n))*(1 - q**(4*l+2*n-2*i))*(1 - q**(2*n+2*i+2))*(1 - q**(4*l-2*i+2*n+2)))

    return ret

def monic_polynomial_from_three_term(l, n):
    pols = [identity_matrix(SR, 2*l+1)]

    i = 1

    while i <= n:
        Xi = three_term_Xn(l, i-1)
        
        if i == 1:
            pols.append(pols[0]*(x*identity_matrix(SR, 2*l+1) - Xi))
        else:
            Yi = three_term_Yn(l, i-1)
            pols.append(pols[i-1]*(x*identity_matrix(SR, 2*l+1) - Xi) - pols[i-2]*Yi)
        i += 1

    return pols[n]


###
# Three-term recurrence relation Pn (spherical), Corollary 4.12

def three_term_An(l, n):
    ret = matrix(SR, 2*l+1, 2*l+1)
    
    for i in range(2*l+1):
        ret[i,i] = 1/(2*q) * (1 - q**(2*n + 2))*(1 - q**(4*l + 2*n + 4))/((1 - q**(2*i + 2*n + 2))*(1 - q**(4*l - 2*i + 2*n + 2)))

    return ret

def three_term_Bn(l, n):
    ret = matrix(SR, 2*l+1, 2*l+1)

    for i in range(2*l):
        ret[i, i+1] = q**(2*n + 1)/2 * (1 - q**(4*l - 2*i))*(1 - q**(2*i + 2)) / ((1 - q**(4*l + 2*n - 2*i))*(1 - q**(2*i + 2*n + 4))) 
    
    for i in range(1, 2*l+1):
        ret[i, i-1] = q**(2*n + 1)/2 * (1 - q**(2*i))*(1 - q**(4*l - 2*i + 2))/((1 - q**(2*n + 2*i))*(1 - q**(4*l - 2*i + 2*n + 4)))

    return ret

def three_term_Cn(l, n):
    ret = matrix(SR, 2*l+1, 2*l+1)

    for i in range(2*l+1):
        ret[i,i] = q/2 * (1 - q**(2*n))*(1 - q**(4*l + 2*n + 2)) / ((1 - q**(2*n + 2*i + 2))*(1 - q**(4*l - 2*i + 2*n + 2)))

    return ret

def polynomial_from_three_term(l, n):
    pols = [identity_matrix(SR, 2*l+1)]
    i = 1

    while i <= n:
        Ai = three_term_An(l, i-1)
        Bi = three_term_Bn(l, i-1)

        if i == 1:
            pols.append(pols[0]*(x*Ai**(-1) - Bi*Ai**(-1)))
        else:
            Ci = three_term_Cn(l, i-1)
            pols.append(pols[i-1]*(x*Ai**(-1) - Bi*Ai**(-1)) - pols[i-2]*Ci*Ai**(-1))

        i += 1

    return pols[n]


###
# q-Difference equations, Theorem 4.13

def M1(l):
    Mret = matrix(SR, 2*l+1, 2*l+1)
    
    for i in range(2*l):
        Mret[i,i+1] = -q**(1 - i) * (1 - q**(2*i + 2)) / (1 - q**2)**2 \
            * z / (1 - z**2)

    for i in range(2*l + 1):
        Mret[i,i] = q**(1 - i) / (1 - q**2)**2 \
            * (1 - q**(2*i + 2)*z**2) / (1 - z**2)

    return Mret

def N1(l):
    Nret = matrix(SR, 2*l+1, 2*l+1)

    for i in range(2*l):
        Nret[i,i+1] = q**(1 - i) * (1 - q**(2*i + 2)) / (1 - q**2)**2 \
            * z / (1 - z**2)
    
    for i in range(2*l + 1):
        Nret[i,i] = q**(1 - i) / (1 - q**2)**2 \
            * (1 - q**(2*i + 2)*z**(-2)) / (1 - z**(-2))
        #    * (1 - q**(2*i + 2)*z**(-2)) / (1 - z**(-2))

    return Nret

def M2(l):
    Mret = matrix(SR, 2*l+1, 2*l+1)

    for i in range(1, 2*l+1):
        Mret[i,i-1] = -q**(1-2*l+i) * (1 - q**(4*l-2*i+2))/((1 - q**2)**2) * z / (1 - z**2)

    for i in range(2*l+1):
        Mret[i,i] = q**(1-2*l+i)/((1 - q**2)**2) * (1 - q**(4*l-2*i+2)*z**2)/(1 - z**2)

    return Mret
    #return antimatrix(SR, 2*l+1) * M1(l) * antimatrix(SR, 2*l+1)

def N2(l):
    return antimatrix(SR, 2*l+1) * N1(l) * antimatrix(SR, 2*l+1)

def D1(l, poly):
    p1 = M1(l)*poly.apply_map(replace_x_z).substitute({z : q*z})
    p2 = N1(l)*poly.apply_map(replace_x_z).substitute({z : q**(-1)*z})

    return SimplifyMatrix(p1 + p2).apply_map(replace_z_x)

def D2(l, poly):
    p1 = M2(l)*poly.apply_map(replace_x_z).substitute({z : q*z})
    p2 = N2(l)*poly.apply_map(replace_x_z).substitute({z : q**(-1)*z})

    return SimplifyMatrix(p1 + p2).apply_map(replace_z_x)

def Lambda1(l, d):
    mret = matrix(SR, 2*l+1)
    
    for i in range(2*l+1):
        mret[i,i] = (q**(-i-d-1) + q**(i+d+1))/((q**(-1)-q)**2)

    return mret

def Lambda2(l, d):
    mret = matrix(SR, 2*l+1)

    for i in range(2*l+1):
        mret[i,i] = (q**(-2*l+i-d-1) + q**(2*l-i+d+1))/((q**(-1)-q)**2)

    return mret


###
# LDU decomposition of the weight, Theorem 4.15.

def c(k, l):
    return (1 - q**(4*k + 2)) * qpoch(q**2, q**2, 2*l + k + 1) \
        * qpoch(q**2, q**2, 2*l - k) * qpoch(q**2, q**2, k)**4 \
        / (qpoch(q**2, q**2, 2*k + 1)**2 * qpoch(q**2, q**2, 2*l)**2)

def T(x, l):
    matt = matrix(SR, 2*l+1, 2*l+1)

    for k in range(2*l + 1):
        kweight = replace_z_x(qpoch([z**2, z**(-2)], q**2, k+1))
        matt[k,k] = q**(-2*l) * c(k,l) / 4  * kweight / (1 - x**2)

    return matt

def L(x, l):
    matl = matrix(SR, 2*l+1, 2*l+1)

    for m in range(2*l+1):
        for k in range(m+1):
            matl[m, k] = q**(m - k) * qpoch(q**2, q**2, m) * qpoch(q**2, q**2, 2*k + 1) \
                / (qpoch(q**2, q**2, m+k+1) * qpoch(q**2, q**2, k)) \
                * q_ultraspherical(m - k, x, q**(2*k + 2), q**2)

    return matl

def Linv(x, l):
    matlinv = matrix(SR, 2*l+1)

    for k in range(2*l+1):
        for n in range(2*l+1):
            matlinv[k, n] = q**((2*k+1)*(k-n)) * qpoch(q**2, q**2, k) * qpoch(q**2, q**2, k+n) \
                / (qpoch(q**2, q**2, 2*k) * qpoch(q**2, q**2, n)) \
                * q_ultraspherical(k-n, x, q**(-2*k), q**2)

    return matlinv

def weight_ldu(l):
    matl = L(x, l)
    matt = T(x, l)

    return SimplifyMatrix(matl * matt * transpose(matl))


###
# Block decomposition of the weight

def Y(l):
    mret = matrix(SR, 2*l+1, 2*l+1)

    if (2*l)%2 == 0:
        for i in range(l):
            mret[i,i] = 1
            mret[2*l-i, 2*l-i] = 1
            mret[i, 2*l-i] = 1
            mret[2*l-i, i] = -1
        mret[l, l] = sqrt(2)
    else:
        for i in range(l+1/2):
            mret[i,i] = 1
            mret[2*l-i, 2*l-i] = 1
            mret[i, 2*l-i] = 1
            mret[2*l-i, i] = -1

    return mret

def weight_block(l):
    return Y(l)**(-1)*weight_ldu(l)*Y(l)


###
# Polynomials explicit in terms of q-Racah and continuous q-ultraspherical 
# polynomials, Theorem 4.18.

def Rn(l, d):
    return transpose(L(x, l))*monic_polynomial_from_three_term(l, d)

def Rn_explicit(l, n):
    matr = matrix(SR, 2*l+1)

    for i, j in itertools.product(range(2*l+1), repeat=2):
        if n + j >= i:
            matr[i,j] = (-1)**i * 2**(-n) * q**(j*(2*i+1) + 2*i*(2*l+n+1) - i**2) \
                * qpoch([q**2, q**(4*l+4)], q**2, n) / qpoch([q**(2*j+2), q**(4*l-2*j+2)], q**2, n) \
                * qpoch([q**(-4*l), q**(-2*j-2*n)], q**2, i) / qpoch([q**2, q**(4*l+4)], q**2, i) \
                * qpoch(q**2, q**2, n+j-i) / qpoch(q**(4*i+4), q**2, n+j-i) \
                * q_Racah(i, j, 1, 1, q**(-2*n-2*j-2), q**(-4*l-2), q**2) \
                * q_ultraspherical(n+j-i, x, q**(2*i+2), q**2)

    return matr

def polynomial_from_qracah_qultra(l, n):
    mret = matrix(SR, 2*l+1, 2*l+1)

    for i, j in itertools.product(range(2*l+1), repeat=2):
        for k in range(i, 2*l+1):
            if n+j >= k:
                mret[i, j] += (-1)**k * q**(n + (2*k+1)*(k-i) + j*(2*k+1) + 2*k*(2*l+n+1) - k**2) \
                    * qpoch(q**2, q**2, k) * qpoch(q**2, q**2, k+i) / (qpoch(q**2, q**2, 2*k) * qpoch(q**2, q**2, i)) \
                    * qpoch([q**(-4*l), q**(-2*j-2*n)], q**2, k) / qpoch([q**2, q**(4*l+4)], q**2, k) \
                    * qpoch(q**2, q**2, n+j-k) / qpoch(q**(4*k+4), q**2, n+j-k) \
                    * q_Racah(k, j, 1, 1, q**(-2*n-2*j-2), q**(-4*l-2), q**2) \
                    * q_ultraspherical(k-i, x, q**(-2*k), q**2) * q_ultraspherical(n+j-k, x, q**(2*k+2), q**2)

    return mret


###
# q-Difference equation for the Rn polynomials, Theorem 7.1.

def K1(l,z):
    rmat = matrix(SR, 2*l+1, 2*l+1)

    for i in range(2*l+1):
        rmat[i,i] = q**(1-i)/((1 - q**2)**2) * (1 - q**(2*i+2)*z**2)/(1 - z**2)

    return rmat

def K2(l,z):
    rmat = matrix(SR, 2*l+1, 2*l+1)


    for i in range(1, 2*l+1):
        rmat[i,i-1] = -q**(i-2*l+1) * (1 - q**(4*l-2*i+2))/((1 - q**2)**2) * z/(1 - z**2)

    for i in range(2*l+1):
        rmat[i,i] = 2*q**(i-2*l+1) * 1/((1 - q**2)**2) * (1 + q**(4*l+2))/((1 + q**(2*i))*(1 + q**(2*i+2))) * (1 - q**(2*i+2)*z**2)/(1 - z**2)

    for i in range(2*l):
        rmat[i,i+1] = -q**(i-2*l+1) * 1/((1 - q**2)**2) * (1 - q**(4*l+2*i+4))*(1 - q**(2*i+2))**2 / ((1 - q**(4*i+6))*(1 - q**(4*i+2))*(1 + q**(2*i+2))**2) * (1 - q**(2*i+2)*z**2)*(1 - q**(2*i+4)*z**2)/(z*(1 - z**2))

    return rmat

def RnD1(l, poly):
    p1 = K1(l,z)*poly.apply_map(replace_x_z).substitute({z : q*z})
    p2 = K1(l,z**(-1))*poly.apply_map(replace_x_z).substitute({z : q**(-1)*z})

    return SimplifyMatrix(p1 + p2).apply_map(replace_z_x)

def RnD2(l, poly):
    p1 = K2(l,z)*poly.apply_map(replace_x_z).substitute({z : q*z})
    p2 = K2(l,z**(-1))*poly.apply_map(replace_x_z).substitute({z : q**(-1)*z})

    return SimplifyMatrix(p1 + p2).apply_map(replace_z_x)


# Full check of all Theorems for lower dimensional polynomials
if __name__ == '__main__':
    import sys

    print 'This is the autochecker for most Theorems of the article \'Matrix-valued orthogonal polynomials related to the quantum analogue of (SU(2) x SU(2), SU(2))\' written by Noud Aldenhoven, Erik Koelink and Pablo Roman.'

    dimension = 4
    num_of_polynomials = 3

    print 'We check all polynomials P^l_n(x), where 1 <= 2*l+1 <= {0} and 0 <= n <= {1}'.format(dimension, num_of_polynomials-1)

    # We skip the spherical polynomials since they take a lot of work to 
    # compute. The polynomials are checked on a super computer for
    # l = 0, 1/2, ..., 9/2, 5 and n = 0, 1, ..., 10.
    spherical_polynomials = []
    print 'Compute polynomials from the full spherical functions. (This may take a while)'
    for l in srange(0, dimension/2, 1/2):
        print 'Computing polynomials of dimension l = {0:4}'.format(str(l)),
        polynomials = []
        for n in range(num_of_polynomials):
            polynomials.append(polynomial_from_spherical(l, n))
            sys.stdout.write('.')
            sys.stdout.flush()
        spherical_polynomials.append(polynomials)
        print

    three_term_polynomials = []
    print 'Compute polynomials from the three term recurrence relation.'
    for l in srange(0, dimension/2, 1/2):
        print 'Computing polynomials of dimension l = {0:4}'.format(str(l)),
        polynomials = []
        for n in range(num_of_polynomials):
            polynomials.append(polynomial_from_three_term(l, n))
            sys.stdout.write('.')
            sys.stdout.flush()
        three_term_polynomials.append(polynomials)
        print

    direct_polynomials = []
    print 'Compute polynomials from the explicit expression in terms of q-Racah and q-ultraspherical polynomials.'
    for l in srange(0, dimension/2, 1/2):
        print 'Computing polynomials of dimension l = {0:4}'.format(str(l)),
        polynomials = []
        for n in range(num_of_polynomials):
            polynomials.append(polynomial_from_qracah_qultra(l, n))
            sys.stdout.write('.')
            sys.stdout.flush()
        direct_polynomials.append(polynomials)
        print
    
    print 'Check that the spherical polynomials and three term recurrence polynomials are equal.',
    are_equal = True
    for pols1, pols2 in zip(three_term_polynomials, spherical_polynomials):
        sys.stdout.write('.')
        sys.stdout.flush()
        for p1, p2 in zip(pols1, pols2):
            if SimplifyMatrix(p1) != SimplifyMatrix(p2):
                print 'False!'
                print three_term_polynomials.index(pols1), pols1.index(p1)
                are_equal = False
                break
    if are_equal:
        print 'True!'

    print 'Check that the three term recurrence polynomials and direct polynomials are equal.',
    are_equal = True
    for pols1, pols2 in zip(three_term_polynomials, direct_polynomials):
        sys.stdout.write('.')
        sys.stdout.flush()
        for p1, p2 in zip(pols1, pols2):
            if SimplifyMatrix(p1) != SimplifyMatrix(p2):
                print 'False!'
                print three_term_polynomials.index(pols1), pols1.index(p1)
                are_equal = False
                break
    if are_equal:
        print 'True!'

    print 'Check orthogonality of the polynomials for the explicit weight and the LDU decomposition of the weight. (This may take a while)'
    for l in srange(0, dimension/2, 1/2):
        print 'Check for dimension l = {0:4}'.format(str(l)),
        are_orthogonal = True
        for m in range(num_of_polynomials):
            for n in range(m, num_of_polynomials):
                sys.stdout.write('.')
                sys.stdout.flush()

                pol1 = three_term_polynomials[2*l][m]
                pol2 = three_term_polynomials[2*l][n]
                rint_direct = matrix_inproduct(pol1, pol2, weight_direct(l))
                rint_ldu = matrix_inproduct(pol1, pol2, weight_ldu(l))

                if m != n:
                    if SimplifyMatrix(rint_direct) != zero_matrix(SR, 2*l+1, 2*l+1) or SimplifyMatrix(rint_ldu) != zero_matrix(SR, 2*l+1, 2*l+1):
                        print 'False!'
                        break
                elif SimplifyMatrix(rint_direct) != SimplifyMatrix(Gn(l,m)) or SimplifyMatrix(rint_ldu) != SimplifyMatrix(Gn(l,m)):
                    print 'False!'
                    break
        if are_orthogonal:
            print 'True!'

    print 'Check the q-Difference equations for the polynomials. (This also takes some time, but not as much as the previous checks)'
    for l in srange(0, dimension/2, 1/2):
        print 'Check for dimension l = {0:4}'.format(str(l)),
        are_qdiff = True
        for n in range(num_of_polynomials):
            sys.stdout.write('.')
            sys.stdout.flush()

            pol = three_term_polynomials[2*l][n]
            d1p = pol**(-1)*D1(l, pol)
            d2p = pol**(-1)*D2(l, pol)

            if SimplifyMatrix(d1p) != SimplifyMatrix(Lambda1(l, n)) or SimplifyMatrix(d2p) != SimplifyMatrix(Lambda2(l, n)):
                print 'False!'
                are_qdiff = False
                break
        if are_qdiff:
            print 'True!'
            
