'''qpolynomials.sage - Created by Noud Aldenhoven (2014).'''

load('bhs.sage')
load('complextools.sage')


## continuous quantum polynomials

def monic_askey_wilson(n, x, a, b, c, d, q, replace_z=True):
    if n < 0:
        return 0
    elif n == 0:
        return 1
    
    Ak = lambda k : (1 - a*b*q**k)*(1 - a*c*q**k)*(1 - a*d*q**k)*(1 - a*b*c*d*q**(k-1)) \
        / (a*(1 - a*b*c*d*q**(2*k-1))*(1 - a*b*c*d*q**(2*k)))
    Ck = lambda k : a*(1 - q**k)*(1 - b*c*q**(k-1))*(1 - b*d*q**(k-1))*(1 - c*d*q**(k-1)) \
        / ((1 - a*b*c*d*q**(2*k-2))*(1 - a*b*c*d*q**(2*k-1)))

    return (x - 1/2*(a + a**(-1) - Ak(n-1) - Ck(n-1))) * monic_askey_wilson(n-1, x, a, b, c, d, q) \
        - 1/4 * Ak(n-2)*Ck(n-1) * monic_askey_wilson(n-2, x, a, b, c, d, q)

def askey_wilson(n, x, a, b, c, d, q, replace_z=True):
    '''askey_wilson(n, z, a, b, c, d, q, replace_z=True): Askey-Wilson polynomials.'''
    lc = 2**n * qpoch(a*b*c*d*q**(n-1), q, n)
    return lc * monic_askey_wilson(n, x, a, b, c, d, q)


def continuous_q_hahn(n, z, a, b, c, q, replace_z=True):
    '''continuous_q_hahn(n, z, a, b, c, q, replace_z=True): Continuous q-Hahn polynomials.'''

    return askey_wilson(n, z, a, b, c, 0, q, replace_z)


def big_q_jacobi(n, x, a, b, c, q):
    '''big_q_jacobi(n, x, a, b, c, q): Big q-Jacobi polynomials.'''

    return bhs([q**(-n), a*b*q**(n+1), x], [a*q, c*q], q, q)


def big_q_legendre(n, x, c, q):
    '''big_q_legendre(n, x, c, q): Big q-Legendre polynomials.'''

    return big_q_jacobi(n, 1, 1, c, q)


def al_salam_chihara(n, z, a, b, q, replace_z=True):
    '''al_salam_chihara(n, z, a, b, q, replace_z=True): Al-Salam-Chihara polynomials.'''

    return continuous_q_hahn(n, z, a, b, 0, q, replace_z)


def q_ultraspherical(n, x, b, q, replace_z=True):
    '''q_ultraspherical(n, x, b, q, replace_z=True): continuous q-Ultraspherical polynomials.'''

    if n < 0:
        return 0

    ret = 0
    for k in range(n+1):
        ret += qpoch(b, q, k) * qpoch(b, q, n-k) \
            / (qpoch(q, q, k) * qpoch(q, q, n-k)) * z**(n - 2*k)

    if replace_z:
        return replace_z_x(ret)
    else:
        return ret

rogers = q_ultraspherical


def monic_continuous_q_Jacobi(n, x, a, b, q):
    if n < 0:
        return 0
    elif n == 0:
        return 1

    Ak = lambda k : q**(-a/2-1/4) * (1 - q**(k+a+1))*(1 - q**(k+a+b+1))*(1 + q**(k+1/2*(a+b+1)))*(1 + q**(k+1/2*(a+b+2))) \
        / ((1 - q**(2*k+a+b+1))*(1 - q**(2*k+a+b+2)))
    Ck = lambda k : q**(a/2+1/4) * (1 - q**k)*(1 - q**(k+b))*(1 + q**(k+1/2*(a+b)))*(1 + q**(k+1/2*(a+b+1))) \
        / ((1 - q**(2*k+a+b))*(1 - q**(2*k+a+b+1)))

    return (x - 1/2*(q**(a/2+1/4) + q**(-a/2-1/4) - Ak(n-1) - Ck(n-1))) * monic_continuous_q_Jacobi(n-1, x, a, b, q) \
        - 1/4 * Ak(n-2)*Ck(n-1) * monic_continuous_q_Jacobi(n-2, x, a, b, q)

def continuous_q_Jacobi(n, x, a, b, q):
    lc = 2**n * q**((a/2+1/4)*n) * qpoch(q**(n+a+b+1), q, n) / qpoch([q, -q**(1/2*(a+b+1)), -q**(1/2*(a+b+2))], q, n)
    return lc * monic_continuous_q_Jacobi(n, x, a, b, q)

def old_continuous_q_Jacobi(n, x, a, b, q, replace_z=True):
    '''continuous_q_Jacobi(n, x, a, b, q, replace_z=True): continuous q-Jacobi polynomials.'''
    
    lc = qpoch(q**(a + 1), q, n) / qpoch(q, q, n)
    poly = bhs(
        [q**(-n), q**(n + a + b + 1), q**(1/2*a + 1/4)*z, 
            q**(1/2*a + 1/4)*z**(-1)],
        [q**(a + 1), -q**(1/2*(a + b + 1)), q**(1/2*(a + b + 2))], q, q)

    if replace_z:
        return lc*replace_z_x(poly)
    else:
        return lc*poly


## Infinite discrete quantum polynomials

def little_q_Jacobi(n, x, a, b, q):
    '''little_q_Jacobi(n, x, a, b, q): little q-Jacobi polynomials.'''

    bhs = BasicHypergeometricSeries([q**(-n), a*b*q**(n+1)], [a*q], q, q*x)
    return bhs.evaluate()



## Discrete quantum polynomials

def q_Racah(n, x, a, b, c, d, q):
    '''q_Racah(n, x, a, b, c, d, q, replace_z=True): q-Racah polynomials.'''

    M = var('M')
    def _get_solution(term):
        sols = solve(term, M, solution_dict=True)
        if len(sols) == 1:
            return sols[0][M] if sols[0][M] > 0 else infinity
        else:
            return infinity

    N = min(map(_get_solution, 
        [q**(-M) == a*q, q**(-M) == b*d*q, q**(-M) == c*q]))

    if n > N:
        return 0

    poly = bhs(
        [q**(-n), a*b*q**(n+1), q**(-x), c*d*q**(x+1)],
        [a*q, b*d*q, c*q], q, q
    )

    return poly

def dual_q_Hahn(n, x, c, d, N, q):
    '''dual_q_Hahn(n, x, c, d, N, q): dual q-Hahn polynomials.'''

    if n != N:
        l1 = [q**(-n), q**(-x), c*d*q**(x + 1)]
        l2 = [c*q, q**(-N)]
        result = BasicHypergeometricSeries(l1, l2, q, q).evaluate()
    else:
        # For n = N we must understand the dual q-Hahn polynomials as
        # polynomials by continuity.
        result = 0
        for k in range(N + 1):
            qpoch1 = qPochhammerSymbol([q**(-x), c*d*q**(x + 1)], q, k)
            if qpoch1.evaluate() == 0:
                break
            qpoch2 = qPochhammerSymbol([c*q, q], q, k)
            result += qpoch1.evaluate() / qpoch2.evaluate() * q**k

    return result       


